// COPYRIGHT: FUNDACIÓN TECNALIA RESEARCH & INNOVATION, 2022.

const faker = require("@faker-js/faker")
const inquirer = require("inquirer")

const {
  createRandomLocation,
  promptAssetSelection,
  TYPES,
  STATUSES
} = require("@hypercog/utils")

const createAsset = async (apiClient, { assetName }) => {
  const newAsset = await apiClient.createAsset({
    id: `asst_${faker.datatype.uuid()}`,
    type: TYPES.RESIDUO_PROCESOS_TERMICOS,
    units: "kg",
    quantity: faker.datatype.number({ min: 400, max: 5000 }),
    fields: {
      description: "Escoria",
      name: assetName,
      status: STATUSES.SLAG,
      castingId: 1
    },
    location: createRandomLocation()
  })

  console.log("Registered asset", newAsset.id)

  return newAsset.id
}

const editComposition = async (
  apiClient,
  { assetId: paramAssetId, caoPercentage }
) => {
  const assetId =
    paramAssetId || (await promptAssetSelection(await apiClient.richQuery()))

  let chosenCaoPercentage = caoPercentage
  if (!caoPercentage) {
    const { inputCaoPercentage } = await inquirer.prompt([
      {
        type: "number",
        name: "inputCaoPercentage",
        message:
          "Choose the percentage of CaO content in the slag (min: 40%, max 70%):",
        default: faker.datatype.number({ min: 40, max: 70 }),
        validate: n => {
          if (n >= 40 && n <= 70) return true
          return "The CaO percentage must be between 40 and 70"
        }
      }
    ])
    chosenCaoPercentage = inputCaoPercentage
  }

  const previousVersion = await apiClient.getAsset(assetId)

  await apiClient.modifyAsset({
    ...previousVersion,
    id: assetId,
    fields: {
      ...previousVersion.fields,
      composition: { cao: chosenCaoPercentage }
    }
  })

  console.log("New composition measure for", assetId)
}

const updateBatchRange = (batch = {}, individualSlag = {}) => {
  const { cao = null } = (individualSlag.fields || {}).composition || {}
  const { cao: caoRange = null } = (batch.fields || {}).composition || {}

  if (cao === null) {
    return caoRange
  }

  if (caoRange === null) {
    return { min: cao, max: cao }
  }

  return {
    min: cao < caoRange.min ? cao : caoRange.min,
    max: cao > caoRange.max ? cao : caoRange.max
  }
}

const sendToCone = async (apiClient, { assetId: paramAssetId }) => {
  const assetId =
    paramAssetId ||
    (await promptAssetSelection(
      await apiClient.richQuery({
        fields: {
          status: STATUSES.SLAG
        }
      })
    ))

  const assetToBeSent = await apiClient.getAsset(assetId)
  const [firstAsset = false] = await apiClient.richQuery({
    fields: {
      status: STATUSES.SLAG_BATCH
    }
  })

  const newConeQuantity =
    assetToBeSent.quantity + (firstAsset ? firstAsset.quantity : 0)

  const coneAsset = await apiClient.joinAsset(
    //`cone_${faker.datatype.uuid()}`,
    firstAsset ? [firstAsset.id, assetId] : [assetId],
    { parentsShouldExist: false, bidirectional: true },
    {
      type: TYPES.RESIDUO_HIERRO_ACERO,
      location: firstAsset ? firstAsset.location : assetToBeSent.location,
      units: "kg",
      quantity: newConeQuantity,
      fields: {
        name: "Slag batch (in cone)",
        status: STATUSES.SLAG_BATCH,
        composition: {
          cao: updateBatchRange(firstAsset, assetToBeSent)
        }
      }
    }
  )

  // Archive individual slag and previous cone
  await apiClient.deleteAsset(assetId)
  if (firstAsset) await apiClient.deleteAsset(firstAsset.id)

  console.log("Moved to cone", coneAsset.id)

  return coneAsset.id
}

const updateStockRange = (stock = {}, batch = {}) => {
  const { cao: caoRange = null } = (stock.fields || {}).composition || {}
  const { cao: newCaoRange = null } = (batch.fields || {}).composition || {}

  if (caoRange === null) {
    return newCaoRange
  }

  if (newCaoRange === null) {
    return caoRange
  }

  return {
    min: newCaoRange.min < caoRange.min ? newCaoRange.min : caoRange.min,
    max: newCaoRange.max > caoRange.max ? newCaoRange.max : caoRange.max
  }
}

const sendToStock = async (
  apiClient,
  { assetId: paramAssetId, stockId: paramStockId }
) => {
  const assetId =
    paramAssetId ||
    (await promptAssetSelection(
      await apiClient.richQuery({
        fields: {
          status: STATUSES.SLAG_BATCH
        }
      })
    ))

  const assetToBeSent = await apiClient.getAsset(assetId)
  if (assetToBeSent.fields.status !== STATUSES.SLAG_BATCH) {
    console.error("You can only combine a cone with a stock")
    return
  }

  const selectedStockId =
    paramStockId ||
    (await promptAssetSelection(
      [
        ...(await apiClient.richQuery({
          fields: {
            status: STATUSES.STOCK
          }
        })),
        {
          id: "newStock",
          type: TYPES.RESIDUO_HIERRO_ACERO,
          fields: { name: "New stock", status: STATUSES.STOCK }
        }
      ],
      "Select the stock where the cone will be merged"
    ))

  if (!selectedStockId) {
    console.error("You must select a stock")
    return
  }

  let previousStock = undefined
  if (selectedStockId !== "newStock") {
    previousStock = await apiClient.getAsset(selectedStockId)
  }
  const newStockQuantity =
    assetToBeSent.quantity + (previousStock ? previousStock.quantity : 0)

  const newStockAsset = await apiClient.joinAsset(
    selectedStockId === "newStock" ? [assetId] : [selectedStockId, assetId],
    { parentsShouldExist: true, bidirectional: true },
    {
      type: TYPES.RESIDUO_HIERRO_ACERO,
      location:
        selectedStockId === "newStock"
          ? assetToBeSent.location
          : (
              await apiClient.getAsset(selectedStockId)
            ).location,
      units: "kg",
      quantity: newStockQuantity,
      fields: {
        status: STATUSES.STOCK,
        name: "Stock",
        composition: {
          cao: updateStockRange(previousStock, assetToBeSent)
        }
      }
    }
  )

  // Archive individual slag and previous cone
  await apiClient.deleteAsset(assetId)
  if (selectedStockId !== "newStock")
    await apiClient.deleteAsset(selectedStockId)

  console.log("Moved to stock", newStockAsset.id)

  return newStockAsset
}

const bidResponse =
  action =>
  async (apiClient, { assetId: paramAssetId, bidder: paramBidder }) => {
    const assetId =
      paramAssetId ||
      (await promptAssetSelection(
        await apiClient.richQuery({
          fields: {
            status: STATUSES.STOCK
          }
        })
      ))

    const assetSelected = await apiClient.getAsset(assetId)
    if (!assetSelected.fields.bids || assetSelected.fields.bids.length === 0) {
      console.error(`There are no active bids to be ${action}ed`)
      return
    }

    let bidder = paramBidder
    if (!bidder) {
      const { inputBidder } = await inquirer.prompt([
        {
          type: "list",
          name: "inputBidder",
          message: `Select the bidder whose bid will be ${action}ed`,
          choices: assetSelected.fields.bids.map(b => ({
            name: `${b.bidder.id} (quantity: ${b.quantity} ${assetSelected.units}, price: ${b.price} €)`,
            value: b.bidder
          }))
        }
      ])
      bidder = inputBidder
    }

    const sentAsset = await apiClient[`${action}Bid`](assetId, bidder.id)
    console.log(
      `Bid from ${bidder.id} ${action}ed for asset ${assetId} (sent ${sentAsset.id})`
    )

    return sentAsset.id
  }

const acceptBid = bidResponse("accept")
const rejectBid = bidResponse("reject")

const discardStock = async (apiClient, { assetId: paramAssetId, quantity }) => {
  const discadableAssets = await apiClient.richQuery({
    fields: {
      status: STATUSES.STOCK
    }
  })

  const assetId = paramAssetId || (await promptAssetSelection(discadableAssets))
  const assetToDiscard = discadableAssets.find(({ id }) => id === assetId)

  let chosenQuantity = quantity
  if (!quantity) {
    const { inputQuantity } = await inquirer.prompt([
      {
        type: "number",
        name: "inputQuantity",
        message: `Choose an amount of slag to discard (max: ${assetToDiscard.quantity} ${assetToDiscard.units}):`,
        default: Math.floor(assetToDiscard.quantity * 0.1) // Default: 10%
      }
    ])
    chosenQuantity = inputQuantity
  }

  const [newStock, discardedBatch] = await apiClient.deleteAsset(
    assetId,
    chosenQuantity
  )

  console.log(`Discarded stock ${discardedBatch} (new stock ${newStock})`)
  return newStock
}

// To be reused as a library in @hypercog/batch-sim
module.exports = {
  createAsset,
  editComposition,
  sendToCone,
  sendToStock,
  acceptBid,
  rejectBid,
  discardStock
}
