// COPYRIGHT: FUNDACIÓN TECNALIA RESEARCH & INNOVATION, 2021.

require("dotenv").config()

const { Command, Option } = require("commander")

const inquirer = require("inquirer")
const faker = require("@faker-js/faker")
faker.locale = "es"

const HypercogApiClient = require("./client")
const { version } = require("./package.json")

/*
10	RESIDUOS DE PROCESOS TÉRMICOS
10 02	Residuos de la industria del hierro y del acero
10 02 01	Residuos del tratamiento de escorias
10 02 02	Escorias no tratadas
10 02 07*	Residuos sólidos del tratamiento de gases que contienen sustancias peligrosas
10 02 08	Residuos sólidos del tratamiento de gases, distintos de los especificados en el código 10 02 07
10 02 10	Cascarilla de laminación
10 02 11*	Residuos del tratamiento del agua de refrigeración que contienen aceites
10 02 12	Residuos del tratamiento del agua de refrigeración distintos de los especificados en el código 10 02 11
10 02 13*	Lodos y tortas de filtración del tratamiento de gases que contienen sustancias peligrosas
10 02 14	Lodos y tortas de filtración del tratamiento de gases, distintos de los especificados en el código 10 02 13
10 02 15	Otros lodos y tortas de filtración
10 02 99	Residuos no especificados en otra categoría
*/

const TYPES = {
  RESIDUO_PROCESOS_TERMICOS: "10",
  RESIDUO_HIERRO_ACERO: "10 02",
  RESIDUOS_TRATAMIENTO_ESCORIAS: "10 02 01",
  ESCORIAS_NO_TRATADAS: "10 02 02",
  RESIDUOS_GASES_PELIGROSOS: "10 02 07",
  RESIDUOS_GASES_OTROS: "10 02 08",
  CASCARILLA_LAMINACION: "10 02 10",
  RESIDUOS_AGUA_ACEITES: "10 02 11",
  RESIDUOS_AGUA_OTROS: "10 02 12",
  LODOS_TORTAS_GASES_PELIGROSOS: "10 02 13",
  LODOS_TORTAS_GASES_OTROS: "10 02 14",
  OTROS_LODOS_TORTAS: "10 02 15",
  RESIDUOS_OTROS: "10 02 99"
}

const STATUSES = {
  SLAG: "slag",
  SLAG_BATCH: "batched", // escoria en cono
  STOCK: "stocked",
  SOLD: "sold",
  DISCARDED: "discarded"
}

const createCLIApp = () => {
  const program = new Command()
  program
    .version(version)
    .addOption(new Option("-a, --api <value>", "API URL").env("API_URL"))
    .addOption(
      new Option(
        "-n, --network <value>",
        "Network the API should connect to"
      ).env("NETWORK")
    )
    .addOption(
      new Option("-c, --channel <value>", "Channel name").env("CHANNEL_NAME")
    )
    .addOption(
      new Option("-h, --chaincode <value>", "Chaincode name").env(
        "CHAINCODE_NAME"
      )
    )
    .addOption(
      new Option(
        "-u, --user <value>",
        "User who will operate with blockchain. E.g., user@organization.com"
      ).env("USER_ID") // User is already an env variable in my OS
    )
    .addOption(
      new Option("-p, --password <value>", "Password for the user").env(
        "PASSWORD"
      )
    )
  return program
}

const loginAndCallAfterwards = fn => async (arg1, cmd) => {
  const { api, network, channel, chaincode, user, password } = cmd.parent.opts()

  const apiClient = new HypercogApiClient(api, network, channel, chaincode)

  const [username, organization] = user.split("@")

  const okLogin = await apiClient.signIn(username, password, organization)
  if (!okLogin) {
    console.error("Unauthorized login")
    return
  }

  fn(apiClient, cmd.opts())
}

const createRandomLocation = () => ({
  coords: {
    lat: Number(faker.address.latitude(43.4613742, 36.7182857)),
    // For some reason faker.address.longitude(0, -9.1952226) does not seem to work
    lng: Number(
      faker.datatype
        .number({
          max: 0,
          min: -9.1952226,
          precision: parseFloat((0.0).toPrecision(4) + "1")
        })
        .toFixed(4)
    )
  },
  desc: faker.address.streetAddress()
})

const promptAssetSelection = async (assets = [], message = "Select asset") => {
  const choices = assets.map(a => ({
    name: `${a.id} (${a.fields.name}, type: ${a.type}, status: ${a.fields.status})`,
    value: a
  }))

  if (choices.length === 0) {
    throw new Error("Suitable assets not found")
  }

  const { selectedAsset } = await inquirer.prompt([
    {
      type: "list",
      name: "selectedAsset",
      message,
      choices
    }
  ])

  return selectedAsset.id
}

const decodeUserId = userId => {
  const [mspId, rest] = userId.split(":")
  const ascii = Buffer.from(rest, "base64").toString("ascii")
  const result = /^x509::CN=(\w+).*,O=([\w\.\-]+).*/gi.exec(ascii)
  if (result) {
    const [_, username, organization] = result
    return { mspId, username, organization }
  }
}

module.exports = {
  HypercogApiClient,
  createCLIApp,
  loginAndCallAfterwards,
  createRandomLocation,
  promptAssetSelection,
  decodeUserId,
  TYPES,
  STATUSES
}
