/**
 * operations.go
 *
 * COPYRIGHT: FUNDACIÓN TECNALIA RESEARCH & INNOVATION, 2021.
 */

package controller

import (
	"errors"

	"git.code.tecnalia.com/ledgerbuilder/sdk/shared"

	"git.code.tecnalia.com/traceblock/sdk/controller"
	"git.code.tecnalia.com/traceblock/sdk/controller/asset"
	"git.code.tecnalia.com/traceblock/sdk/controller/barcode"
	"git.code.tecnalia.com/traceblock/sdk/controller/join"
	"git.code.tecnalia.com/traceblock/sdk/controller/qr"
	"git.code.tecnalia.com/traceblock/sdk/controller/split"
	"git.code.tecnalia.com/traceblock/sdk/controller/transfer"
	"git.code.tecnalia.com/traceblock/sdk/middleware"

	"git.code.tecnalia.com/blockchain/hypercog/controller/bid"
	"git.code.tecnalia.com/blockchain/hypercog/controller/discard"
	"git.code.tecnalia.com/blockchain/hypercog/controller/stats"
	m2 "git.code.tecnalia.com/blockchain/hypercog/middleware"
)

var (
	assetController			asset.TraceableAssetController
	// for debug/testing functions
	testController			controller.DprngController
	// for ui related helper calls
	uiController			controller.UIController
	// transfer controller
	transferController		*transfer.TransferAssetController
	splitController			*split.SplitController
	joinController			*join.JoinController
	qrController			*qr.QRController
	barcodeController		*barcode.BarcodeController
	// bid controller
	bidController			*bid.BidController
	bidDecisionController	*bid.BidDecisionController
	
	discardController		*discard.DiscardController
)

func init() {
	assetController = asset.NewTraceableAssetController()
	transferController = transfer.NewTransferAssetController()
	splitController = split.NewSplitController()
	joinController = join.NewJoinController()
	qrController = qr.NewQRController()
	barcodeController = barcode.NewBarcodeController()
	discardController = discard.NewDiscardController()
	// for ui related helper calls
	uiController = controller.NewUIController()
	// for debug/testing functions
	testController = controller.NewHelperController()
	// bid controllers
	bidController = bid.NewBidController()
	bidDecisionController = bid.NewBidDecisionController()
}

// Implementation of chaincode business logic operations
func ContextOperations(m shared.AbstractChaincodeOperationManager) error {
	if m != nil {
		controller.NewOperationsBuilder().
			NewDefaults(&assetController, assetController.StubReader, assetController.LedgerReader).
			AddOperation(
				"asset-register",
				shared.WRITE_OP,
				[]shared.MiddlewareInterface{
					m2.ValidateUnits,
					m2.ValidateStatus,
					m2.ValidateType,
					middleware.InjectAssetIdIfNone,
					middleware.RejectIfAssetExists,
					middleware.MarkAssetAsCreated,
					middleware.InjectAssetOwnerData,
				},
				[]shared.MiddlewareInterface{
					m2.UpdateSlagStats,
				},
				assetController.SaveAbstractAsset,
			).
			AddOperation(
				"asset-read",
				shared.READ_OP,
				[]shared.MiddlewareInterface{
					middleware.RejectIfNoId,
					middleware.RejectIfAssetNotExist,
				},
				nil,
				assetController.OnlySameOrg(assetController.ReadAbstractAsset),
			).
			AddBasicReadOperation(
				"asset-history",
				assetController.OnlySameOrg(assetController.ReadAssetHistory),
			).
			AddOperation(
				"asset-update",
				shared.WRITE_OP,
				[]shared.MiddlewareInterface{
					m2.ValidateUnits,
					m2.ValidateOptionalStatus,
					m2.ValidateOptionalType,
					middleware.RejectIfNoId,
					middleware.RejectIfAssetNotExist,
					middleware.MarkAssetAsModified,
				},
				nil,
				assetController.OnlySameOrgAndRole(assetController.UpdateAsset),
			).
			AddOperationWithoutDefaults(
				"asset-archive",
				shared.WRITE_OP,
				nil,
				discardController.StubReader,
				discardController.LedgerReader,
				nil,
				nil,
				discardController.OnlySameOrgAndRole(discardController.Discard),
			).
			AddOperationWithoutDefaults(
				"asset-transfer",
				shared.WRITE_OP,
				nil,
				transferController.StubReader,
				transferController.LedgerReader,
				nil,
				nil,
				transferController.OnlySameOrgAndRole(transferController.TransferAssetOwnership),
			).
			AddOperationWithoutDefaults(
				"asset-join",
				shared.WRITE_OP,
				&assetController,
				joinController.StubReader,
				joinController.LedgerReader,
				[]shared.MiddlewareInterface{
					m2.ValidateJoinStatus,
					m2.ValidateJoinType,
				},
				nil,
				joinController.JoinSingle,
			).
			AddOperationWithoutDefaults(
				"asset-split",
				shared.WRITE_OP,
				nil,
				splitController.StubReader,
				splitController.LedgerReader,
				[]shared.MiddlewareInterface{
					m2.ValidateSplitBaseStatus,
					m2.ValidateSplitBaseType,
				},
				nil,
				splitController.OnlySameOrgAndRole(splitController.SplitSingle),
			).
			// Dashboard operations
			NewDefaults(nil, uiController.StubReader, uiController.LedgerReader).
			AddBasicReadOperation("dashboard-asset-allowed-states", GetAssetAllowedStates).
			AddBasicReadOperation("dashboard-asset-types-list", GetAssetTypesList).
			AddOperation(
				"asset-owned-elements",
				shared.READ_OP,			
				[]shared.MiddlewareInterface{
					// This should go first to parse incoming asset-type object and return a selector
					middleware.InjectTypeConstraint,
					middleware.InjectOrgConstraint,
					middleware.InjectRoleConstraint,
				},
				nil,
				assetController.RichQuery,
			).
			AddOperation(
				"rich-query",
				shared.READ_OP,			
				[]shared.MiddlewareInterface{
					m2.InjectOrgConstraint,
				},
				nil,
				assetController.RichQuery,
			).
			// QR & barcode operations
			NewDefaults(nil, qrController.StubReader, qrController.LedgerReader).
			AddBasicReadOperation("qr-create", qrController.GenerateAssetQR).
			AddBasicReadOperation("qr-resolve", qrController.ResolveQr).
			NewDefaults(nil, barcodeController.StubReader, barcodeController.LedgerReader).
			AddBasicReadOperation("barcode-create", barcodeController.GenerateAssetBarcode).
			AddBasicReadOperation("barcode-resolve", barcodeController.ResolveBarcode).
			// Other operations
			EmptyDefaults().
			AddBasicReadOperation("random-uuid", testController.DistributedPrngUuid).
			AddBasicReadOperation("random-string", testController.DistributedPrngString).
			// HyperCOG operations
			AddOperation(
				"hypercog-stats",
				shared.READ_OP,			
				[]shared.MiddlewareInterface{
					m2.RejectIfNotPublicAdmin,
				},
				nil,
				stats.GetStatsOperation,
			).
			AddOperation(
				"hypercog-list-bid-assets",
				shared.READ_OP,			
				[]shared.MiddlewareInterface{
					m2.RejectIfNotCementCompany,
					m2.InjectStockedAssetConstraint,
					m2.InjectSidenorOrgConstraint, // Order is important
				},
				nil,
				assetController.RichQuery, // Actually implemented in LowLevelController
			).
			AddOperationWithoutDefaults(
				"hypercog-bid-asset",
				shared.WRITE_OP,
				&assetController, // Not used, just because it is required
				bidController.StubReader,
				bidController.LedgerReader,
				[]shared.MiddlewareInterface{
					m2.RejectIfNotCementCompany,
				},
				nil,
				bidController.OwnedBySidenor(bidController.BidAsset),
			).
			AddOperationWithoutDefaults(
				"hypercog-reject-asset",
				shared.WRITE_OP,
				&assetController, // Not used, just because it is required
				bidDecisionController.StubReader,
				bidDecisionController.LedgerReader,
				[]shared.MiddlewareInterface{
					m2.RejectIfNotSidenor,
				},
				nil,
				assetController.OnlySameOrg(bidDecisionController.RejectBid),
			).
			AddOperationWithoutDefaults(
				"hypercog-accept-asset",
				shared.WRITE_OP,
				&assetController, // Not used, just because it is required
				bidDecisionController.StubReader,
				bidDecisionController.LedgerReader,
				[]shared.MiddlewareInterface{
					m2.RejectIfNotSidenor,
				},
				nil,
				assetController.OnlySameOrg(bidDecisionController.AcceptBid),
			).
			// For debugging stats
			AddOperation(
				"hypercog-stats-reset",
				shared.WRITE_OP,		
				[]shared.MiddlewareInterface{
					m2.RejectIfNotPublicAdmin,
				},
				nil,
				stats.ResetStats,
			).
			Build(m)

		return nil
	}
	return errors.New("no operation manager found")
}
