/**
 * stub.go
 *
 * COPYRIGHT: FUNDACIÓN TECNALIA RESEARCH & INNOVATION, 2022.
 */

package stats

import (
	"encoding/json"
	"fmt"
	"math"

	"git.code.tecnalia.com/ledgerbuilder/sdk/core/util/logging"
	"git.code.tecnalia.com/ledgerbuilder/sdk/shared"

	hmodel "git.code.tecnalia.com/blockchain/hypercog/model"
	"git.code.tecnalia.com/blockchain/hypercog/model/stats"
	"git.code.tecnalia.com/traceblock/sdk/model"
)

const STEEL_ORG_LIST_KEY = "steel_orgs"
const CEMENT_ORG_LIST_KEY = "cement_orgs"

type OrgList []string

var log = logging.NewGoLogger("stats")

func getSteelStats(stub shared.LedgerBuildrStubInterface, key string) (st *stats.RawSteelOrgStats, err error) {
	value, err := stub.GetState(key)

	if err != nil || len(value) == 0 {
		st = new(stats.RawSteelOrgStats)
		st.MinPrice = math.MaxFloat64
	} else {
		err = json.Unmarshal(value, &st)
		if err != nil  {
			return nil, err
		}
	}

	return st, nil
}

func getCementStats(stub shared.LedgerBuildrStubInterface, key string) (st *stats.RawCementOrgStats, err error) {
	value, err := stub.GetState(key)

	if err != nil || len(value) == 0 {
		st = new(stats.RawCementOrgStats)
		st.MinPrice = math.MaxFloat64
	} else {
		err = json.Unmarshal(value, &st)
		if err != nil  {
			return nil, err
		}
	}

	return st, nil
}

func putJson(stub shared.LedgerBuildrStubInterface, key string, stats interface{}) (error) {
	serialized, err := json.Marshal(stats)
	if err != nil {
		return err
	}

	err = stub.PutState(key, serialized)
	if err != nil {
		return err
	}

	return nil
}

func _getOrgList(stub shared.LedgerBuildrStubInterface, key string) (st *OrgList, err error) {
	value, err := stub.GetState(key)
	if err != nil || len(value) == 0 {
		st = new(OrgList)
	} else {
		err = json.Unmarshal(value, &st)
		if err != nil  {
			return nil, err
		}
	}

	return st, nil
}

func getSteelOrgList(stub shared.LedgerBuildrStubInterface) (st *OrgList, err error) {
	return _getOrgList(stub, STEEL_ORG_LIST_KEY)
}

func getCementOrgList(stub shared.LedgerBuildrStubInterface) (st *OrgList, err error) {
	return _getOrgList(stub, CEMENT_ORG_LIST_KEY)
}

func resetStats(stub shared.LedgerBuildrStubInterface) (error) {
	orgs, err := getSteelOrgList(stub)
	if len(*orgs) > 0 {
		for _, org := range *orgs {
			err = stub.DelState(org)
			if err != nil {
				return err
			}
		}
	}

	orgs, err = getCementOrgList(stub)
	if len(*orgs) > 0 {
		for _, org := range *orgs {
			err = stub.DelState(org)
			if err != nil {
				return err
			}
		}
	}

	err = stub.DelState(STEEL_ORG_LIST_KEY)
	err = stub.DelState(CEMENT_ORG_LIST_KEY)

	return err
}

func (orgs OrgList) contains(org string) (bool) {
	for _, o := range orgs {
        if o == org {
            return true
        }
    }
	return false
}

func addOrgIfNotExist(stub shared.LedgerBuildrStubInterface, key string, org string) (err error) {
	orgs, err := _getOrgList(stub, key)
	if err != nil {
		return err
	}

	if !orgs.contains(org) {
		newOrgs := append(*orgs, org)

		serialized, err := json.Marshal(newOrgs)
		if err != nil {
			return err
		}
	
		err = stub.PutState(key, serialized)
		if err != nil {
			return err
		}

		log.Debug("Added stat for new org")
	}

	return nil
}


func retrieveExistingStats(stub shared.LedgerBuildrStubInterface) (*stats.RawSteelOrgStats, error) {
	org, err := model.GetOrganization(stub)
	if err != nil {
		return nil, err
	}

	customStats, err := getSteelStats(stub, org)
	if err != nil {
		return nil, err
	}

	return customStats, nil
}

func upgradeSteelStats(stub shared.LedgerBuildrStubInterface, customStats *stats.RawSteelOrgStats) (error) {
	org, err := model.GetOrganization(stub)
	if err != nil {
		return err
	}

	err = putJson(stub, org, customStats)
	if err != nil {
		return err
	}

	err = addOrgIfNotExist(stub, STEEL_ORG_LIST_KEY, org)
	if err != nil {
		return err
	}

	log.Debug("Upgrading stats")
	return nil
}

func upgradeCementStats(stub shared.LedgerBuildrStubInterface, customStats *stats.RawCementOrgStats, org string) (error) {
	err := putJson(stub, org, customStats)
	if err != nil {
		return err
	}

	err = addOrgIfNotExist(stub, CEMENT_ORG_LIST_KEY, org)
	if err != nil {
		return err
	}

	log.Debug("Upgrading stats")
	return nil
}

func RegisterSlag(stub shared.LedgerBuildrStubInterface, quantity uint32, units string) (error) {
	orgStats, err := retrieveExistingStats(stub)
	if err != nil {
		return err
	}

	orgStats.RegisterSlag(quantity, units)

	err = upgradeSteelStats(stub, orgStats)
	if err != nil {
		return err
	}

	return nil
}


func RegisterDiscard(stub shared.LedgerBuildrStubInterface, quantity uint32, units string) (error) {
	orgStats, err := retrieveExistingStats(stub)
	if err != nil {
		return err
	}

	orgStats.RegisterDiscard(quantity, units)

	err = upgradeSteelStats(stub, orgStats)
	if err != nil {
		return err
	}

	return nil
}

func RegisterSale(stub shared.LedgerBuildrStubInterface, acceptedBid *hmodel.Bid, units string) (error) {
	senderOrgStats, err := retrieveExistingStats(stub)
	if err != nil {
		return err
	}

	senderOrgStats.UpgradeSale(acceptedBid.Quantity, acceptedBid.Price, units)

	err = upgradeSteelStats(stub, senderOrgStats)
	if err != nil {
		return err
	}


	receiverOrgStats, err := getCementStats(stub, acceptedBid.Bidder.Org)
	if err != nil {
		return err
	}

	receiverOrgStats.UpgradeSale(acceptedBid.Quantity, acceptedBid.Price, units)
		
	err = upgradeCementStats(stub, receiverOrgStats, acceptedBid.Bidder.Org)
	if err != nil {
		return err
	}


	log.Debug(fmt.Sprintf("Sale registered in stats (quantity: %d %s, price: %.2f €)", acceptedBid.Quantity, units, acceptedBid.Price))
	return nil
}