/**
 * status.go
 *
 * COPYRIGHT: FUNDACIÓN TECNALIA RESEARCH & INNOVATION, 2021.
 */

package middleware

import (
	"errors"

	"git.code.tecnalia.com/blockchain/hypercog/constants"
	"git.code.tecnalia.com/ledgerbuilder/sdk/shared"
	"git.code.tecnalia.com/traceblock/sdk/controller/join"
	"git.code.tecnalia.com/traceblock/sdk/controller/split"
	"git.code.tecnalia.com/traceblock/sdk/middleware"
)

var (
	ValidateStatus = shared.NewMiddlewareFunction("validate-status", _rejectIfUnsetOrInvalidStatus)
	ValidateOptionalStatus = shared.NewMiddlewareFunction("validate-status-optional", _rejectIfInvalidStatus)
	ValidateSplitBaseStatus = shared.NewMiddlewareFunction("validate-split-base-status", _rejectIfInvalidSplitBaseStatus)
	ValidateJoinStatus = shared.NewMiddlewareFunction("validate-join-status", _rejectIfInvalidJoinStatus)

	errInvalidStatus = errors.New("invalid asset status")
)

func containsValidStatus(arbitraryDataFields map[string]interface{}, rejectIfEmpty bool) error {
	field, ok := arbitraryDataFields["status"].(string)
	if ok && constants.IsValidStatus(field) {
		return nil
	}

	if !rejectIfEmpty && (!ok || field == "") {
		return nil
	}

	return errInvalidStatus
}

func _checkStatus(requestAsset shared.LedgerBuildrAsset, rejectIfEmpty bool) (shared.LedgerBuildrAsset, error) {
	traceableAsset, err := middleware.ConvertToTraceableAsset(requestAsset)
	if err != nil {
		return nil, err
	}

	err = containsValidStatus(traceableAsset.ArbitraryDataFields, rejectIfEmpty)
	if err != nil {
		return nil, err
	}

	return requestAsset, nil
}

func _rejectIfUnsetOrInvalidStatus(stub shared.LedgerBuildrStubInterface, ctl shared.ControllerInterface, req shared.TXRequestInterface, requestAsset shared.LedgerBuildrAsset) (shared.LedgerBuildrAsset, error) {
	return _checkStatus(requestAsset, true)
}

func _rejectIfInvalidStatus(stub shared.LedgerBuildrStubInterface, ctl shared.ControllerInterface, req shared.TXRequestInterface, requestAsset shared.LedgerBuildrAsset) (shared.LedgerBuildrAsset, error) {
	return _checkStatus(requestAsset, false)
}

func _rejectIfInvalidSplitBaseStatus(stub shared.LedgerBuildrStubInterface, ctl shared.ControllerInterface, req shared.TXRequestInterface, requestAsset shared.LedgerBuildrAsset) (shared.LedgerBuildrAsset, error) {
	if requestAsset == nil {
		return nil, errInvalidStatus
	}

	params, ok := requestAsset.(*split.SplitParams)
	if !ok {
		return nil, errInvalidStatus
	}

	err := containsValidStatus(params.SplitConfig.BaseAsset.ArbitraryDataFields, true)
	if err != nil {
		return nil, err
	}

	return requestAsset, nil
}

func _rejectIfInvalidJoinStatus(stub shared.LedgerBuildrStubInterface, ctl shared.ControllerInterface, req shared.TXRequestInterface, requestAsset shared.LedgerBuildrAsset) (shared.LedgerBuildrAsset, error) {
	if requestAsset == nil {
		return nil, errInvalidStatus
	}

	params, ok := requestAsset.(*join.JoinParams)
	if !ok {
		return nil, errInvalidStatus
	}

	err := containsValidStatus(params.ArbitraryDataFields, true)
	if err != nil {
		return nil, err
	}

	return requestAsset, nil
}