/**
 * type.go
 *
 * COPYRIGHT: FUNDACIÓN TECNALIA RESEARCH & INNOVATION, 2021.
 */

package middleware

import (
	"errors"

	"git.code.tecnalia.com/blockchain/hypercog/constants"
	"git.code.tecnalia.com/ledgerbuilder/sdk/shared"
	"git.code.tecnalia.com/traceblock/sdk/controller/join"
	"git.code.tecnalia.com/traceblock/sdk/controller/split"
	"git.code.tecnalia.com/traceblock/sdk/middleware"
)

var (
	ValidateType = shared.NewMiddlewareFunction("validate-type", _rejectIfUnsetOrInvalidType)
	ValidateOptionalType = shared.NewMiddlewareFunction("validate-type-optional", _rejectIfInvalidType)
	ValidateSplitBaseType = shared.NewMiddlewareFunction("validate-split-base-status", _rejectIfInvalidSplitBaseType)
	ValidateJoinType = shared.NewMiddlewareFunction("validate-join-status", _rejectIfInvalidJoinType)

	errInvalidType = errors.New("invalid asset type")
)

func containsValidType(assetType string, rejectIfEmpty bool) error {
	if !rejectIfEmpty && assetType == "" {
		return nil
	}

	if constants.Exists(assetType) {
		return nil
	}
	
	return errInvalidType
}

// https://eur-lex.europa.eu/legal-content/EN/TXT/HTML/?uri=CELEX:32014D0955&from=DE
func _checkType(requestAsset shared.LedgerBuildrAsset, rejectIfEmpty bool) (shared.LedgerBuildrAsset, error) {
	asst, err := middleware.ConvertToTraceableAsset(requestAsset)
	if err != nil {
		return nil, err
	}

	asst.AssetType = constants.NormalizeCode(asst.AssetType)
	err = containsValidType(asst.AssetType, rejectIfEmpty)
	if err != nil {
		return nil, err
	}
	
	return requestAsset, nil
}

func _rejectIfUnsetOrInvalidType(stub shared.LedgerBuildrStubInterface, ctl shared.ControllerInterface, req shared.TXRequestInterface, requestAsset shared.LedgerBuildrAsset) (shared.LedgerBuildrAsset, error) {
	return _checkType(requestAsset, true)
}

func _rejectIfInvalidType(stub shared.LedgerBuildrStubInterface, ctl shared.ControllerInterface, req shared.TXRequestInterface, requestAsset shared.LedgerBuildrAsset) (shared.LedgerBuildrAsset, error) {
	return _checkType(requestAsset, false)
}

func _rejectIfInvalidSplitBaseType(stub shared.LedgerBuildrStubInterface, ctl shared.ControllerInterface, req shared.TXRequestInterface, requestAsset shared.LedgerBuildrAsset) (shared.LedgerBuildrAsset, error) {
	if requestAsset == nil {
		return nil, errInvalidStatus
	}

	params, ok := requestAsset.(*split.SplitParams)
	if !ok {
		return nil, errInvalidStatus
	}

	params.SplitConfig.BaseAsset.AssetType = constants.NormalizeCode(params.SplitConfig.BaseAsset.AssetType)
	err := containsValidType(params.SplitConfig.BaseAsset.AssetType, true)
	if err != nil {
		return nil, err
	}

	return requestAsset, nil
}

func _rejectIfInvalidJoinType(stub shared.LedgerBuildrStubInterface, ctl shared.ControllerInterface, req shared.TXRequestInterface, requestAsset shared.LedgerBuildrAsset) (shared.LedgerBuildrAsset, error) {
	if requestAsset == nil {
		return nil, errInvalidStatus
	}

	params, ok := requestAsset.(*join.JoinParams)
	if !ok {
		return nil, errInvalidStatus
	}

	params.AssetType = constants.NormalizeCode(params.AssetType)
	err := containsValidType(params.AssetType, true)
	if err != nil {
		return nil, err
	}

	return requestAsset, nil
}