// COPYRIGHT: FUNDACIÓN TECNALIA RESEARCH & INNOVATION, 2021.

require("dotenv").config()

const ApiClient = require("@traceblock/api-client")
const SDKHelper = require("./sdk")

const usersByOrg = require("./users")

const readPassword = (user, org) =>
  process.env[`PASSWD-${user}-${org.replace(/\./g, "-")}`.toUpperCase()] ||
  "secret"

const modifyUser = async (
  client,
  username,
  password,
  organization,
  role,
  email,
  affiliation
) => {
  const oldCert = await client.getUser(username, organization)
  oldCert.attrs = [
    ...oldCert.attrs,
    { name: "affiliation", value: affiliation, ecert: true }
  ]

  const modUser = await client.modifyUser(
    username,
    password,
    organization,
    role,
    email,
    oldCert
  )

  if (!modUser) {
    throw new Error(`Invalid ${role} modification`)
  }

  return modUser
}

const createOrModifyUserIfNeeded = async (
  client,
  username,
  password,
  organization,
  role,
  affiliation
) => {
  const email = `${username}@${organization}`

  try {
    // Await is needed here to force the catch clausule if something goes wrong
    return await client.createUser(
      username,
      password,
      organization,
      role,
      email,
      "en",
      [{ name: "affiliation", value: affiliation, ecert: true }]
    )
  } catch (err) {
    return modifyUser(
      client,
      username,
      password,
      organization,
      role,
      email,
      affiliation
    )
  }
}

const printAffiliations = async (sdkClient, organization) => {
  console.log(`\nAffiliations for ${organization}`)

  const existingAffiliations = await sdkClient.getAffiliations(organization)
  existingAffiliations.forEach(({ name, affiliations: affs = [] }) =>
    console.log(`\t${name}: [ ${affs.map(a => a.name).join(", ")} ] `)
  )
}

const createAffiliations = async (
  sdkClient,
  organization,
  affiliations = []
) => {
  if (affiliations.length > 0) {
    for (let affiliation of affiliations) {
      try {
        await sdkClient.addAffiliation(organization, affiliation)
      } catch (err) {
        err.errors.forEach(({ message }) => console.error("Error:", message))
      }
    }

    await printAffiliations(sdkClient, organization)
  }
}

const createUsers = async (useSdk = false) => {
  let client = new ApiClient(
    process.env.TRACEBLOCK_API,
    process.env.TRACEBLOCK_NETWORK,
    process.env.TRACEBLOCK_CHANNEL,
    process.env.TRACEBLOCK_CHAINCODE
  )

  if (useSdk) {
    // CA nodes must be accesible.
    // If this script is not run from a container inside the network, expose the ports.
    // Append "--expose-endpoints true" to deployment/network/up.sh
    client = new SDKHelper(process.env.CONNECTION_CONFIG)
  }

  for (let el of Object.entries(usersByOrg)) {
    const [org, values] = el
    const admin = values.find(v => v.role === "admin")

    await client.signIn(admin.user, readPassword(admin.user, org), org)

    if (useSdk) {
      // The API does not include an option to create new affiliations
      await createAffiliations(
        client,
        org,
        values.filter(u => !!u.affiliation).map(u => u.affiliation)
      )
    }

    const prom = values
      //.filter(v => v.role !== "admin")
      // Admin roles also need to be configured
      .map(u =>
        createOrModifyUserIfNeeded(
          client,
          u.user,
          readPassword(u.user, org),
          org,
          u.role,
          u.affiliation
        )
      )

    await Promise.all(prom)
  }

  console.log("Users created")
}

const [clientType = ""] = process.argv.slice(2)
createUsers(clientType.toLowerCase() === "sdk")
