// COPYRIGHT: FUNDACIÓN TECNALIA RESEARCH & INNOVATION, 2022.

require("dotenv").config()

const fs = require("fs")
const { Wallets } = require("fabric-network")
const FabricCAServices = require("fabric-ca-client")

const loadProfile = connectionConfigPath => {
  if (fs.existsSync(connectionConfigPath)) {
    return JSON.parse(fs.readFileSync(connectionConfigPath, "utf8"))
  }
  throw new Error(`no such file or directory: ${connectionConfigPath}`)
}

const updateAttrs = (newAttrs = [], oldAttrs = []) => [
  ...oldAttrs
    // Reserved prefix, Registrar is not authorized to register it
    .filter(attr => !attr.name.startsWith("hf."))
    // Remove old values of the new attrs to be added
    .filter(attr => !newAttrs.map(a => a.name).find(e => e === attr.name)),
  ...newAttrs
]

const createAttrs = (organization, role, affiliation, email, lang = "es") => [
  // Attrs used in https://git.code.tecnalia.com/ledgerbuilder/sdk/-/blob/develop/core/stub/stub.go
  { name: "org", value: organization, ecert: true },
  { name: "role", value: role, ecert: true }, // Also used in frontend
  { name: "email", value: email },
  { name: "lang", value: lang, ecert: true }
]

// By default Fabric only defines these 3 affiliations:
// https://hyperledger-fabric-ca.readthedocs.io/en/release-1.4/deployguide/ca-config.html#affiliations
// Since I didn't learn how to modify them in Minifab,
// new affiliations must be prefixed by one of those making it ugly and confusing :(
const toValidHfAffiliation = affiliation =>
  affiliation ? "org1.department1." + affiliation : "org1.department1"

const extractAffiliationFromAttrs = attrs => {
  const affiliationAttr = attrs.find(attr => attr.name === "affiliation")
  if (!!affiliationAttr) {
    return toValidHfAffiliation(affiliationAttr.value)
  }
  // return undefined
}

class SDKHelper {
  constructor(connectionProfilePath) {
    this.connectionProfile = loadProfile(connectionProfilePath)
    this.caClient = null
    this.wallet = {}
  }

  async _getWallet(org) {
    // setup the wallet to hold the credentials of the application user
    if (!this.wallet[org]) {
      this.wallet[org] = await Wallets.newInMemoryWallet()
    }

    return this.wallet[org]
  }

  _getMspId(org) {
    const orgProfile = this.connectionProfile.organizations[org]
    if (!orgProfile) {
      throw new Error(`no such organization in the connection profile: ${org}`)
    }
    return orgProfile.mspid
  }

  _setCAClient(org) {
    const orgProfile = this.connectionProfile.organizations[org]
    if (!orgProfile) {
      throw new Error(`no such organization in the connection profile: ${org}`)
    }

    const caHostName = orgProfile.certificateAuthorities[0]
    const caInfo = this.connectionProfile.certificateAuthorities[caHostName]
    const caTLSCACerts = caInfo.tlsCACerts.pem
    this.caClient = new FabricCAServices(
      caInfo.url,
      { trustedRoots: caTLSCACerts, verify: false },
      caInfo.caName
    )
  }

  async _getAdminUser(organization) {
    const wallet = await this._getWallet(organization)

    const adminIdentity = await wallet.get("admin")
    if (!adminIdentity) {
      throw new Error(
        "An identity for the admin user does not exist in the wallet"
      )
    }

    const provider = wallet
      .getProviderRegistry()
      .getProvider(adminIdentity.type)
    return provider.getUserContext(adminIdentity, "admin")
  }

  async _enrollUser(userId, password, org) {
    const mspId = this._getMspId(org)
    const enrollment = await this.caClient.enroll({
      enrollmentID: userId,
      enrollmentSecret: password
    })
    // x509Identity
    return {
      credentials: {
        certificate: enrollment.certificate,
        privateKey: enrollment.key.toBytes()
      },
      mspId: mspId,
      type: "X.509"
    }
  }

  async signIn(userId, password, org) {
    const wallet = await this._getWallet(org)

    // Check to see if we've already enrolled the admin user.
    const identity = await wallet.get(userId)
    if (identity) {
      console.log(
        `Signing: An identity for the user ${userId} already exists in the wallet`
      )
      return
    }

    this._setCAClient(org)

    const x509Identity = await this._enrollUser(userId, password, org)
    await wallet.put(userId, x509Identity)
  }

  // Register the user, enroll the user, and import the new identity into the wallet.
  // if affiliation is specified by client, the affiliation value must be configured in CA
  async createUser(
    userId,
    password,
    organization,
    role,
    email,
    language,
    extraAttrs = []
  ) {
    const adminUser = await this._getAdminUser(organization)

    // returns secret
    await this.caClient.register(
      {
        enrollmentID: userId,
        enrollmentSecret: password,
        role: "client", // "hf.Type" in the certificate
        affiliation: extractAffiliationFromAttrs(extraAttrs),
        maxEnrollments: -1, // unlimited
        attrs: updateAttrs(
          createAttrs(organization, role, email, language),
          extraAttrs
        )
      },
      adminUser
    )

    await this._enrollUser(userId, password, organization)
  }

  /**
   * Modifies the information of a user.
   * Note that only a user with an `admin` role can do this.
   *
   * @param {string} username Username of the user.
   * @param {string} password New password of the user.
   * @param {string} organization New organization of the user.
   * @param {string} role New role for the user.
   * @param {string} email E-mail of the new user.
   * @param {object} [oldCert] X.509 certificate fields
   * @param {Array.<AttributeX509>} [oldCert.attrs=[]] Attributes of the certificate.
   */
  async modifyUser(userId, password, organization, role, email, oldCert = {}) {
    const wallet = await this._getWallet(organization)

    // Must use an admin to register a new user
    const adminIdentity = await wallet.get("admin")
    if (!adminIdentity) {
      console.log("An identity for the admin user does not exist in the wallet")
      console.log("Enroll the admin user before retrying")
      return
    }

    const provider = wallet
      .getProviderRegistry()
      .getProvider(adminIdentity.type)
    const adminUser = await provider.getUserContext(adminIdentity, "admin")
    const identityService = await this.caClient.newIdentityService()

    return identityService.update(
      userId,
      {
        enrollmentID: userId,
        secret: password,
        affiliation: extractAffiliationFromAttrs(oldCert.attrs),
        role: "client", // "hf.Type" in the certificate
        attrs: updateAttrs(
          createAttrs(organization, role, email),
          oldCert.attrs
        )
      },
      adminUser
    )
  }

  // Order important since "organization" param is not in API client
  async getUser(userId, organization) {
    const identityService = await this.caClient.newIdentityService()
    const { success, result } = await identityService.getOne(
      userId,
      await this._getAdminUser(organization)
    )
    return success ? result : null
  }

  async getAffiliations(organization) {
    const affiliationService = await this.caClient.newAffiliationService()
    const affResp = await affiliationService.getAll(
      await this._getAdminUser(organization)
    )

    if (!affResp.success) {
      console.error(affResp.errors)
      return []
    }

    return affResp.result.affiliations
  }

  async updateAffiliation(organization, oldAffiliation, affiliation) {
    const affiliationService = await this.caClient.newAffiliationService()
    await affiliationService.update(
      oldAffiliation,
      // Force: If any of the parent affiliations do not exist, create all parent affiliations also.
      { name: toValidHfAffiliation(affiliation), force: true },
      await this._getAdminUser(organization)
    )
  }

  async addAffiliation(organization, affiliation) {
    const affiliationService = await this.caClient.newAffiliationService()
    await affiliationService.create(
      // Force: If any of the parent affiliations do not exist, create all parent affiliations also.
      { name: toValidHfAffiliation(affiliation), force: true },
      await this._getAdminUser(organization)
    )
  }
}

module.exports = SDKHelper
